import { IncomingMessage, ServerResponse, RequestListener } from 'http';
import { constants } from 'http2';
import { pluralize } from 'inflection';
import { BaseSchema, ObjectSchema } from 'valibot';
import { SerializerPair } from './serializers';
import {
	handleCreateItem,
	handleGetCollection,
	handleGetItem,
	handleGetRoot,
	handleHasMethodAndUrl
} from './handlers';

export interface DataSource<T = object> {
	initialize(): Promise<unknown>;
	getMultiple(): Promise<T[]>;
	getSingle(id: string): Promise<T | null>;
	create(data: Partial<T>): Promise<T>;
	delete(id: string): Promise<unknown>;
	emplace(id: string, data: Partial<T>): Promise<T>;
	patch(id: string, data: Partial<T>): Promise<T | null>;
}

export interface ApplicationParams {
	name: string;
	dataSource?: (resource: Resource) => DataSource;
}

export interface Resource {
	idAttr: string;
	itemName?: string;
	collectionName?: string;
	routeName?: string;
	dataSource: DataSource;
	newId(dataSource: DataSource): string | number | unknown;
	schema: BaseSchema;
}

interface GenerationStrategy {
	(dataSource: DataSource, ...args: unknown[]): Promise<string | number | unknown>;
}

interface IdParams {
	generationStrategy: GenerationStrategy;
}

export const resource = <T extends BaseSchema>(schema: T) => {
	let theIdAttr: string;
	let theItemName: string;
	let theCollectionName: string;
	let theRouteName: string;
	let idGenerationStrategy: GenerationStrategy;
	const fullTextAttrs = new Set<string>();

	return {
		id(newIdAttr: string, params: IdParams) {
			theIdAttr = newIdAttr;
			idGenerationStrategy = params.generationStrategy;
			return this;
		},
		newId(dataSource: DataSource) {
			return idGenerationStrategy(dataSource);
		},
		fullText(fullTextAttr: string) {
			if (
				schema.type === 'object'
				&& (
					schema as unknown as ObjectSchema<
						Record<string, BaseSchema>,
						undefined,
						Record<string, string>
					>
				)
					.entries[fullTextAttr]?.type === 'string'
			) {
				fullTextAttrs.add(fullTextAttr);
				return this;
			}

			throw new Error(`Could not set attribute ${fullTextAttr} as fulltext.`);
		},
		name(n: string) {
			theItemName = n;
			theCollectionName = theCollectionName ?? pluralize(theItemName).toLowerCase();
			theRouteName = theRouteName ?? theCollectionName;
			return this;
		},
		collection(n: string) {
			theCollectionName = n;
			theRouteName = theRouteName ?? theCollectionName;
			return this;
		},
		route(n: string) {
			theRouteName = n;
			return this;
		},
		get idAttr() {
			return theIdAttr;
		},
		get collectionName() {
			return theCollectionName;
		},
		get itemName() {
			return theItemName;
		},
		get routeName() {
			return theRouteName;
		},
		get schema() {
			return schema;
		}
	};
};

interface CreateServerParams {
	baseUrl?: string;
	host?: string;
}

type RequestListenerWithReturn<
	P extends unknown = unknown, Q extends typeof IncomingMessage = typeof IncomingMessage, R extends typeof ServerResponse = typeof ServerResponse> = (
	...args: Parameters<RequestListener<Q, R>>
) => P;

interface HandlerState {
	handled: boolean;
}

interface ApplicationState {
	resources: Set<Resource>;
	serializers: Map<string, SerializerPair>;
}

interface MiddlewareArgs {
	handlerState: HandlerState;
	appState: ApplicationState;
	appParams: ApplicationParams;
	serverParams: CreateServerParams;
}

export interface Middleware {
	(args: MiddlewareArgs): RequestListenerWithReturn<HandlerState | Promise<HandlerState>>
}

export const application = (appParams: ApplicationParams) => {
	const applicationState: ApplicationState = {
		resources: new Set<Resource>(),
		serializers: new Map<string, SerializerPair>()
	};

	return {
		contentType(mimeTypePrefix: string, serializerPair: SerializerPair) {
			applicationState.serializers.set(mimeTypePrefix, serializerPair);
			return this;
		},
		resource(res: Partial<Resource>) {
			res.dataSource = res.dataSource ?? appParams.dataSource?.(res as Resource);
			if (typeof res.dataSource === 'undefined') {
				throw new Error(`Resource ${res.itemName} must have a data source.`);
			}
			applicationState.resources.add(res as Resource);
			return this;
		},
		async createServer(serverParams = {} as CreateServerParams) {
			const serverModule = await import('http');
			const server = serverModule.createServer();

			server.on('request', async (req, res) => {
				const middlewareState = await [
					handleHasMethodAndUrl,
					handleGetRoot,
					handleGetCollection,
					handleGetItem,
					handleCreateItem,
				]
					.reduce(
						async (currentHandlerStatePromise, middleware) => {
							const currentHandlerState = await currentHandlerStatePromise;
							if (currentHandlerState.handled) {
								return currentHandlerState;
							}

							return middleware({
								handlerState: currentHandlerState,
								appState: applicationState,
								appParams,
								serverParams
							})(req, res);
						},
						Promise.resolve<HandlerState>({
							handled: false
						})
					);

				if (middlewareState.handled) {
					return;
				}

				res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
				res.statusMessage = 'URL Not Found';
				res.end();
			});

			return server;
		}
	};
};
