import { constants } from 'http2';
import Negotiator from 'negotiator';
import { ValiError } from 'valibot';
import { Middleware } from './core';
import { getBody, getMethod, getUrl } from './utils';

export const handleHasMethodAndUrl: Middleware = ({}) => (req, res) => {
	if (!req.method) {
		res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
			'Allow': 'HEAD,GET,POST,PUT,PATCH,DELETE' // TODO check with resources on allowed methods
		});
		res.end();
		return {
			handled: true
		};
	}

	if (!req.url) {
		res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
		res.end();
		return {
			handled: true
		};
	}

	return {
		handled: false
	};
};

export const handleGetRoot: Middleware = ({
	appState,
	appParams,
	serverParams
}) => (req, res) => {
	const method = getMethod(req);
	if (method !== 'GET') {
		return {
			handled: false
		};
	}

	const baseUrl = serverParams.baseUrl ?? '';
	const { url } = getUrl(req, baseUrl);
	if (url !== '/') {
		return {
			handled: false
		};
	}

	const negotiator = new Negotiator(req);
	const availableMediaTypes = Array.from(appState.serializers.keys());
	const theMediaType = negotiator.mediaType(availableMediaTypes);
	if (typeof theMediaType === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
		res.end();
		return {
			handled: true
		};
	}

	const theSerializerPair = appState.serializers.get(theMediaType);
	if (typeof theSerializerPair === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
		res.end();
		return {
			handled: true
		};
	}

	const singleResDatum = {
		name: appParams.name
	};
	const theFormatted = theSerializerPair.serialize(singleResDatum);
	res.writeHead(constants.HTTP_STATUS_OK, {
		'Content-Type': theMediaType,
		// we are using custom headers for links because the standard Link header
		// is referring to the document metadata (e.g. author, next page, etc)
		// https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Link
		'X-Resource-Link': Array.from(appState.resources)
			.map((r) =>
					`<${baseUrl}/${r.routeName}>; name="${r.collectionName}"`,
				// TODO add host?
			)
			.join(', ')
	});
	res.end(theFormatted);
	return {
		handled: true
	};
};

export const handleGetCollection: Middleware = ({
	appState,
	serverParams
}) => async (req, res) => {
	const method = getMethod(req);
	if (method !== 'GET') {
		return {
			handled: false
		};
	}

	const baseUrl = serverParams.baseUrl ?? '';
	const { url } = getUrl(req, baseUrl);

	const [, mainResourceRouteName, mainResourceId = ''] = url.split('/');
	if (mainResourceId !== '') {
		return {
			handled: false
		}
	}

	const theResource = Array.from(appState.resources).find((r) => r.routeName === mainResourceRouteName);
	if (typeof theResource === 'undefined') {
		return {
			handled: false
		};
	}

	const negotiator = new Negotiator(req);
	const availableMediaTypes = Array.from(appState.serializers.keys());
	const theMediaType = negotiator.mediaType(availableMediaTypes);
	if (typeof theMediaType === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
		res.end();
		return {
			handled: true
		};
	}

	const theSerializerPair = appState.serializers.get(theMediaType);
	if (typeof theSerializerPair === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
		res.end();
		return {
			handled: true
		};
	}

	await theResource.dataSource.initialize();
	const resData = await theResource.dataSource.getMultiple(); // TODO paginated responses per resource
	const theFormatted = theSerializerPair.serialize(resData);

	res.writeHead(constants.HTTP_STATUS_OK, {
		'Content-Type': theMediaType,
		'X-Resource-Total-Item-Count': resData.length
	});
	res.end(theFormatted);

	return {
		handled: true
	};
};

export const handleGetItem: Middleware = ({
	appState,
	serverParams
}) => async (req, res) => {
	const method = getMethod(req);
	if (method !== 'GET') {
		return {
			handled: false
		};
	}

	const baseUrl = serverParams.baseUrl ?? '';
	const { url } = getUrl(req, baseUrl);

	const [, mainResourceRouteName, mainResourceId = ''] = url.split('/');
	if (mainResourceId === '') {
		return {
			handled: false
		}
	}

	const theResource = Array.from(appState.resources).find((r) => r.routeName === mainResourceRouteName);
	if (typeof theResource === 'undefined') {
		return {
			handled: false
		};
	}

	const negotiator = new Negotiator(req);
	const availableMediaTypes = Array.from(appState.serializers.keys());
	const theMediaType = negotiator.mediaType(availableMediaTypes);
	if (typeof theMediaType === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
		res.end();
		return {
			handled: true
		};
	}

	const theSerializerPair = appState.serializers.get(theMediaType);
	if (typeof theSerializerPair === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
		res.end();
		return {
			handled: true
		};
	}

	await theResource.dataSource.initialize();
	const singleResDatum = await theResource.dataSource.getSingle(mainResourceId);
	if (singleResDatum) {
		const theFormatted = theSerializerPair.serialize(singleResDatum);
		res.writeHead(constants.HTTP_STATUS_OK, { 'Content-Type': theMediaType });
		res.end(theFormatted);
		return {
			handled: true
		};
	}

	res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
	res.statusMessage = `${theResource.itemName} Not Found`;
	res.end();
	return {
		handled: true
	};
};

export const handleCreateItem: Middleware = ({
	appState,
	serverParams
}) => async (req, res) => {
	const method = getMethod(req);
	if (method !== 'POST') {
		return {
			handled: false
		};
	}

	const baseUrl = serverParams.baseUrl ?? '';
	const { url } = getUrl(req, baseUrl);

	const [, mainResourceRouteName, mainResourceId = ''] = url.split('/');
	if (mainResourceId !== '') {
		return {
			handled: false
		}
	}

	const theResource = Array.from(appState.resources).find((r) => r.routeName === mainResourceRouteName);
	if (typeof theResource === 'undefined') {
		return {
			handled: false
		};
	}

	const theDeserializer = appState.serializers.get(req.headers['content-type'] ?? 'application/octet-stream');
	if (typeof theDeserializer === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE;
		res.end();
		return {
			handled: true
		};
	}

	const negotiator = new Negotiator(req);
	const availableMediaTypes = Array.from(appState.serializers.keys());
	const theMediaType = negotiator.mediaType(availableMediaTypes);
	if (typeof theMediaType === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
		res.end();
		return {
			handled: true
		};
	}

	const theSerializerPair = appState.serializers.get(theMediaType);
	if (typeof theSerializerPair === 'undefined') {
		res.statusCode = constants.HTTP_STATUS_NOT_ACCEPTABLE;
		res.end();
		return {
			handled: true
		};
	}

	await theResource.dataSource.initialize();
	let bodyDeserialized: unknown;
	try {
		bodyDeserialized = await getBody(req, theDeserializer, theResource.schema);
	} catch (errRaw) {
		const err = errRaw as ValiError;
		res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
		res.statusMessage = `Invalid ${theResource.itemName}`;

		if (Array.isArray(err.issues)) {
			// TODO better error reporting, localizable messages
			const theFormatted = theSerializerPair.serialize(
				err.issues.map((i) => (
					`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
				))
			);
			res.end(theFormatted);
		} else {
			res.end();
		}
		return {
			handled: true
		};
	}

	try {
		const newId = await theResource.newId(theResource.dataSource);
		const params = bodyDeserialized as Record<string, unknown>;
		params[theResource.idAttr] = newId;
		const newObject = await theResource.dataSource.create(params);
		const theFormatted = theSerializerPair.serialize(newObject);
		res.writeHead(constants.HTTP_STATUS_OK, {
			'Content-Type': theMediaType,
			'Location': `${serverParams.baseUrl}/${theResource.routeName}/${newId}`
		});
		res.end(theFormatted);
	} catch {
		res.statusCode = constants.HTTP_STATUS_INTERNAL_SERVER_ERROR;
		res.statusMessage = `Could Not Return ${theResource.itemName}`;
		res.end();
	}
	return {
		handled: true
	};
}
